@@ -28,8 +28,7 @@
 /* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                 */
 /**************************************************************************/
 
-#ifndef BVH_H
-#define BVH_H
+#pragma once
 
 // BVH
 // This class provides a wrapper around BVH tree, which contains most of the functionality
@@ -52,6 +51,8 @@
 // and pairable_mask is either 0 if static, or set to all if non static
 
 #include "bvh_tree.h"
+
+#include "core/math/geometry_3d.h"
 #include "core/os/mutex.h"
 
 #define BVHTREE_CLASS BVH_Tree<T, NUM_TREES, 2, MAX_ITEMS, USER_PAIR_TEST_FUNCTION, USER_CULL_TEST_FUNCTION, USE_PAIRS, BOUNDS, POINT>
@@ -405,7 +406,7 @@ public:
 		}
 
 		Vector<Vector3> convex_points = Geometry3D::compute_convex_mesh_points(&p_convex[0], p_convex.size());
-		if (convex_points.size() == 0) {
+		if (convex_points.is_empty()) {
 			return 0;
 		}
 
@@ -435,8 +436,6 @@ private:
 			return;
 		}
 
-		BOUNDS bb;
-
 		typename BVHTREE_CLASS::CullParams params;
 
 		params.result_count_overall = 0;
@@ -770,7 +769,7 @@ private:
 
 	// for collision pairing,
 	// maintain a list of all items moved etc on each frame / tick
-	LocalVector<BVHHandle, uint32_t, true> changed_items;
+	LocalVector<BVHHandle> changed_items;
 	uint32_t _tick = 1; // Start from 1 so items with 0 indicate never updated.
 
 	class BVHLockedFunction {
@@ -806,5 +805,3 @@ public:
 };
 
 #undef BVHTREE_CLASS
-
-#endif // BVH_H
