@@ -28,8 +28,7 @@
 /* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                 */
 /**************************************************************************/
 
-#ifndef REGISTER_CORE_TYPES_H
-#define REGISTER_CORE_TYPES_H
+#pragma once
 
 void register_core_types();
 void register_core_settings();
@@ -38,5 +37,3 @@ void register_early_core_singletons();
 void register_core_singletons();
 void unregister_core_types();
 void unregister_core_extensions();
-
-#endif // REGISTER_CORE_TYPES_H
