@@ -34,19 +34,25 @@
 #include "core/io/dir_access.h"
 #include "core/os/time.h"
 #include "core/version.h"
-#include "editor/editor_paths.h"
-#include "editor/editor_settings.h"
 #include "editor/editor_string_names.h"
-#include "editor/project_manager.h"
+#include "editor/file_system/editor_paths.h"
+#include "editor/project_manager/project_manager.h"
 #include "editor/project_manager/project_tag.h"
+#include "editor/settings/editor_settings.h"
 #include "editor/themes/editor_scale.h"
 #include "scene/gui/button.h"
+#include "scene/gui/dialogs.h"
 #include "scene/gui/label.h"
 #include "scene/gui/line_edit.h"
+#include "scene/gui/progress_bar.h"
 #include "scene/gui/texture_button.h"
 #include "scene/gui/texture_rect.h"
 #include "scene/resources/image_texture.h"
 
+const char *ProjectList::SIGNAL_LIST_CHANGED = "list_changed";
+const char *ProjectList::SIGNAL_SELECTION_CHANGED = "selection_changed";
+const char *ProjectList::SIGNAL_PROJECT_ASK_OPEN = "project_ask_open";
+
 void ProjectListItemControl::_notification(int p_what) {
 	switch (p_what) {
 		case NOTIFICATION_THEME_CHANGED: {
@@ -66,21 +72,60 @@ void ProjectListItemControl::_notification(int p_what) {
 			project_unsupported_features->set_texture(get_editor_theme_icon(SNAME("NodeWarning")));
 
 			favorite_button->set_texture_normal(get_editor_theme_icon(SNAME("Favorites")));
+
 			if (project_is_missing) {
-				explore_button->set_icon(get_editor_theme_icon(SNAME("FileBroken")));
+				explore_button->set_button_icon(get_editor_theme_icon(SNAME("FileBroken")));
+#if !defined(ANDROID_ENABLED) && !defined(WEB_ENABLED)
 			} else {
-				explore_button->set_icon(get_editor_theme_icon(SNAME("Load")));
+				explore_button->set_button_icon(get_editor_theme_icon(SNAME("Load")));
+#endif
 			}
 		} break;
 
 		case NOTIFICATION_MOUSE_ENTER: {
 			is_hovering = true;
 			queue_redraw();
+			queue_accessibility_update();
 		} break;
 
 		case NOTIFICATION_MOUSE_EXIT: {
 			is_hovering = false;
 			queue_redraw();
+			queue_accessibility_update();
+		} break;
+
+		case NOTIFICATION_ACCESSIBILITY_UPDATE: {
+			RID ae = get_accessibility_element();
+			ERR_FAIL_COND(ae.is_null());
+
+			DisplayServer::get_singleton()->accessibility_update_set_role(ae, DisplayServer::AccessibilityRole::ROLE_LIST_BOX_OPTION);
+			DisplayServer::get_singleton()->accessibility_update_set_name(ae, TTR("Project") + " " + project_title->get_text());
+			DisplayServer::get_singleton()->accessibility_update_set_value(ae, project_title->get_text());
+
+			DisplayServer::get_singleton()->accessibility_update_add_action(ae, DisplayServer::AccessibilityAction::ACTION_CLICK, callable_mp(this, &ProjectListItemControl::_accessibility_action_open));
+			DisplayServer::get_singleton()->accessibility_update_add_action(ae, DisplayServer::AccessibilityAction::ACTION_SCROLL_INTO_VIEW, callable_mp(this, &ProjectListItemControl::_accessibility_action_scroll_into_view));
+			DisplayServer::get_singleton()->accessibility_update_add_action(ae, DisplayServer::AccessibilityAction::ACTION_FOCUS, callable_mp(this, &ProjectListItemControl::_accessibility_action_focus));
+			DisplayServer::get_singleton()->accessibility_update_add_action(ae, DisplayServer::AccessibilityAction::ACTION_BLUR, callable_mp(this, &ProjectListItemControl::_accessibility_action_blur));
+
+			ProjectList *pl = get_list();
+			if (pl) {
+				DisplayServer::get_singleton()->accessibility_update_set_list_item_index(ae, pl->get_index(this));
+			}
+			DisplayServer::get_singleton()->accessibility_update_set_list_item_level(ae, 0);
+			DisplayServer::get_singleton()->accessibility_update_set_list_item_selected(ae, is_selected);
+		} break;
+
+		case NOTIFICATION_FOCUS_ENTER: {
+			ProjectList *pl = get_list();
+			if (pl) {
+				int idx = pl->get_index(this);
+				if (idx >= 0) {
+					pl->ensure_project_visible(idx);
+					pl->select_project(idx);
+
+					pl->emit_signal(SNAME(ProjectList::SIGNAL_SELECTION_CHANGED));
+				}
+			}
 		} break;
 
 		case NOTIFICATION_DRAW: {
@@ -88,7 +133,7 @@ void ProjectListItemControl::_notification(int p_what) {
 				draw_style_box(get_theme_stylebox(SNAME("selected"), SNAME("Tree")), Rect2(Point2(), get_size()));
 			}
 			if (is_hovering) {
-				draw_style_box(get_theme_stylebox(SNAME("hover"), SNAME("Tree")), Rect2(Point2(), get_size()));
+				draw_style_box(get_theme_stylebox(SNAME("hovered"), SNAME("Tree")), Rect2(Point2(), get_size()));
 			}
 
 			draw_line(Point2(0, get_size().y + 1), Point2(get_size().x, get_size().y + 1), get_theme_color(SNAME("guide_color"), SNAME("Tree")));
@@ -96,6 +141,53 @@ void ProjectListItemControl::_notification(int p_what) {
 	}
 }
 
+ProjectList *ProjectListItemControl::get_list() const {
+	if (!is_inside_tree()) {
+		return nullptr;
+	}
+	ProjectList *pl = Object::cast_to<ProjectList>(get_parent()->get_parent());
+	return pl;
+}
+
+void ProjectListItemControl::_accessibility_action_scroll_into_view(const Variant &p_data) {
+	ProjectList *pl = get_list();
+	if (pl) {
+		int idx = pl->get_index(this);
+		if (idx >= 0) {
+			pl->ensure_project_visible(idx);
+		}
+	}
+}
+
+void ProjectListItemControl::_accessibility_action_open(const Variant &p_data) {
+	ProjectList *pl = get_list();
+	if (pl && !pl->project_opening_initiated) {
+		pl->emit_signal(SNAME(ProjectList::SIGNAL_PROJECT_ASK_OPEN));
+	}
+}
+
+void ProjectListItemControl::_accessibility_action_focus(const Variant &p_data) {
+	ProjectList *pl = get_list();
+	if (pl) {
+		int idx = pl->get_index(this);
+		if (idx >= 0) {
+			pl->ensure_project_visible(idx);
+			pl->select_project(idx);
+		}
+	}
+}
+
+void ProjectListItemControl::_accessibility_action_blur(const Variant &p_data) {
+	ProjectList *pl = get_list();
+	if (pl) {
+		int idx = pl->get_index(this);
+		if (idx >= 0) {
+			pl->ensure_project_visible(idx);
+			pl->deselect_project(idx);
+		}
+	}
+}
+
 void ProjectListItemControl::_favorite_button_pressed() {
 	emit_signal(SNAME("favorite_pressed"));
 }
@@ -106,10 +198,14 @@ void ProjectListItemControl::_explore_button_pressed() {
 
 void ProjectListItemControl::set_project_title(const String &p_title) {
 	project_title->set_text(p_title);
+	project_title->set_accessibility_name(TTRC("Project Name"));
+	queue_accessibility_update();
 }
 
 void ProjectListItemControl::set_project_path(const String &p_path) {
 	project_path->set_text(p_path);
+	project_path->set_accessibility_name(TTRC("Project Path"));
+	queue_accessibility_update();
 }
 
 void ProjectListItemControl::set_tags(const PackedStringArray &p_tags, ProjectList *p_parent_list) {
@@ -151,7 +247,7 @@ void ProjectListItemControl::set_unsupported_features(PackedStringArray p_featur
 					project_version_major = project_version_split[0].to_int();
 					project_version_minor = project_version_split[1].to_int();
 				}
-				if (VERSION_MAJOR != project_version_major || VERSION_MINOR <= project_version_minor) {
+				if (GODOT_VERSION_MAJOR != project_version_major || GODOT_VERSION_MINOR <= project_version_minor) {
 					// Don't show a warning if the project was last edited in a previous minor version.
 					tooltip_text += TTR("This project was last edited in a different Godot version: ") + p_features[i] + "\n";
 				}
@@ -167,6 +263,7 @@ void ProjectListItemControl::set_unsupported_features(PackedStringArray p_featur
 			return;
 		}
 		project_version->set_tooltip_text(tooltip_text);
+		project_unsupported_features->set_focus_mode(FOCUS_ACCESSIBILITY);
 		project_unsupported_features->set_tooltip_text(tooltip_text);
 		project_unsupported_features->show();
 	} else {
@@ -181,6 +278,7 @@ bool ProjectListItemControl::should_load_project_icon() const {
 void ProjectListItemControl::set_selected(bool p_selected) {
 	is_selected = p_selected;
 	queue_redraw();
+	queue_accessibility_update();
 }
 
 void ProjectListItemControl::set_is_favorite(bool p_favorite) {
@@ -188,22 +286,17 @@ void ProjectListItemControl::set_is_favorite(bool p_favorite) {
 }
 
 void ProjectListItemControl::set_is_missing(bool p_missing) {
-	if (project_is_missing == p_missing) {
-		return;
-	}
 	project_is_missing = p_missing;
 
 	if (project_is_missing) {
 		project_icon->set_modulate(Color(1, 1, 1, 0.5));
 
-		explore_button->set_icon(get_editor_theme_icon(SNAME("FileBroken")));
-		explore_button->set_tooltip_text(TTR("Error: Project is missing on the filesystem."));
+		explore_button->set_button_icon(get_editor_theme_icon(SNAME("FileBroken")));
+		explore_button->set_tooltip_text(TTRC("Error: Project is missing on the filesystem."));
 	} else {
-		project_icon->set_modulate(Color(1, 1, 1, 1.0));
-
-		explore_button->set_icon(get_editor_theme_icon(SNAME("Load")));
 #if !defined(ANDROID_ENABLED) && !defined(WEB_ENABLED)
-		explore_button->set_tooltip_text(TTR("Show in File Manager"));
+		explore_button->set_button_icon(get_editor_theme_icon(SNAME("Load")));
+		explore_button->set_tooltip_text(TTRC("Show in File Manager"));
 #else
 		// Opening the system file manager is not supported on the Android and web editors.
 		explore_button->hide();
@@ -229,6 +322,7 @@ void ProjectListItemControl::_bind_methods() {
 
 ProjectListItemControl::ProjectListItemControl() {
 	set_focus_mode(FocusMode::FOCUS_ALL);
+	set_auto_translate_mode(AUTO_TRANSLATE_MODE_DISABLED);
 
 	VBoxContainer *favorite_box = memnew(VBoxContainer);
 	favorite_box->set_alignment(BoxContainer::ALIGNMENT_CENTER);
@@ -236,6 +330,8 @@ ProjectListItemControl::ProjectListItemControl() {
 
 	favorite_button = memnew(TextureButton);
 	favorite_button->set_name("FavoriteButton");
+	favorite_button->set_tooltip_text(TTRC("Add to favorites"));
+	favorite_button->set_auto_translate_mode(AUTO_TRANSLATE_MODE_ALWAYS);
 	// This makes the project's "hover" style display correctly when hovering the favorite icon.
 	favorite_button->set_mouse_filter(MOUSE_FILTER_PASS);
 	favorite_box->add_child(favorite_button);
@@ -261,7 +357,7 @@ ProjectListItemControl::ProjectListItemControl() {
 		main_vbox->add_child(title_hb);
 
 		project_title = memnew(Label);
-		project_title->set_auto_translate_mode(AUTO_TRANSLATE_MODE_DISABLED);
+		project_title->set_focus_mode(FOCUS_ACCESSIBILITY);
 		project_title->set_name("ProjectName");
 		project_title->set_h_size_flags(Control::SIZE_EXPAND_FILL);
 		project_title->set_clip_text(true);
@@ -283,12 +379,15 @@ ProjectListItemControl::ProjectListItemControl() {
 
 		explore_button = memnew(Button);
 		explore_button->set_name("ExploreButton");
+		explore_button->set_tooltip_auto_translate_mode(AUTO_TRANSLATE_MODE_ALWAYS);
+		explore_button->set_tooltip_text(TTRC("Open in file manager"));
 		explore_button->set_flat(true);
 		path_hb->add_child(explore_button);
 		explore_button->connect(SceneStringName(pressed), callable_mp(this, &ProjectListItemControl::_explore_button_pressed));
 
 		project_path = memnew(Label);
 		project_path->set_name("ProjectPath");
+		project_path->set_focus_mode(FOCUS_ACCESSIBILITY);
 		project_path->set_structured_text_bidi_override(TextServer::STRUCTURED_TEXT_FILE);
 		project_path->set_clip_text(true);
 		project_path->set_h_size_flags(Control::SIZE_EXPAND_FILL);
@@ -302,14 +401,17 @@ ProjectListItemControl::ProjectListItemControl() {
 		project_unsupported_features->hide();
 
 		project_version = memnew(Label);
+		project_version->set_focus_mode(FOCUS_ACCESSIBILITY);
 		project_version->set_name("ProjectVersion");
 		project_version->set_mouse_filter(Control::MOUSE_FILTER_PASS);
 		path_hb->add_child(project_version);
 
 		last_edited_info = memnew(Label);
+		last_edited_info->set_focus_mode(FOCUS_ACCESSIBILITY);
 		last_edited_info->set_name("LastEditedInfo");
 		last_edited_info->set_mouse_filter(Control::MOUSE_FILTER_PASS);
-		last_edited_info->set_tooltip_text(TTR("Last edited timestamp"));
+		last_edited_info->set_tooltip_auto_translate_mode(AUTO_TRANSLATE_MODE_ALWAYS);
+		last_edited_info->set_tooltip_text(TTRC("Last edited timestamp"));
 		last_edited_info->set_modulate(Color(1, 1, 1, 0.5));
 		path_hb->add_child(last_edited_info);
 
@@ -343,22 +445,25 @@ struct ProjectListComparator {
 	}
 };
 
-const char *ProjectList::SIGNAL_LIST_CHANGED = "list_changed";
-const char *ProjectList::SIGNAL_SELECTION_CHANGED = "selection_changed";
-const char *ProjectList::SIGNAL_PROJECT_ASK_OPEN = "project_ask_open";
-
 // Helpers.
 
 bool ProjectList::project_feature_looks_like_version(const String &p_feature) {
-	return p_feature.contains(".") && p_feature.substr(0, 3).is_numeric();
+	return p_feature.contains_char('.') && p_feature.substr(0, 3).is_numeric();
 }
 
 // Notifications.
 
 void ProjectList::_notification(int p_what) {
 	switch (p_what) {
+		case NOTIFICATION_TRANSLATION_CHANGED: {
+			if (is_ready()) {
+				// FIXME: Technically this only needs to update some dynamic texts, not the whole list.
+				update_project_list();
+			}
+		} break;
+
 		case NOTIFICATION_PROCESS: {
-			// Load icons as a coroutine to speed up launch when you have hundreds of projects
+			// Load icons as a coroutine to speed up launch when you have hundreds of projects.
 			if (_icon_load_index < _projects.size()) {
 				Item &item = _projects.write[_icon_load_index];
 				if (item.control->should_load_project_icon()) {
@@ -366,10 +471,68 @@ void ProjectList::_notification(int p_what) {
 				}
 				_icon_load_index++;
 
+				// Scan directories in thread to avoid blocking the window.
+			} else if (scan_data && scan_data->scan_in_progress.is_set()) {
+				// Wait for the thread.
 			} else {
 				set_process(false);
+				if (scan_data) {
+					_scan_finished();
+				}
 			}
 		} break;
+
+		case NOTIFICATION_ACCESSIBILITY_UPDATE: {
+			RID ae = get_accessibility_element();
+			ERR_FAIL_COND(ae.is_null());
+
+			DisplayServer::get_singleton()->accessibility_update_set_role(ae, DisplayServer::AccessibilityRole::ROLE_LIST_BOX);
+			DisplayServer::get_singleton()->accessibility_update_set_list_item_count(ae, _projects.size());
+			DisplayServer::get_singleton()->accessibility_update_set_flag(ae, DisplayServer::AccessibilityFlags::FLAG_MULTISELECTABLE, false);
+		}
+	}
+}
+
+// Projects scan.
+
+void ProjectList::_scan_thread(void *p_scan_data) {
+	ScanData *scan_data = static_cast<ScanData *>(p_scan_data);
+
+	for (const String &base_path : scan_data->paths_to_scan) {
+		print_verbose(vformat("Scanning for projects in \"%s\".", base_path));
+		_scan_folder_recursive(base_path, &scan_data->found_projects, scan_data->scan_in_progress);
+
+		if (!scan_data->scan_in_progress.is_set()) {
+			print_verbose("Scan aborted.");
+			break;
+		}
+	}
+	print_verbose(vformat("Found %d project(s).", scan_data->found_projects.size()));
+	scan_data->scan_in_progress.clear();
+}
+
+void ProjectList::_scan_finished() {
+	if (scan_data->scan_in_progress.is_set()) {
+		// Abort scanning.
+		scan_data->scan_in_progress.clear();
+	}
+
+	scan_data->thread->wait_to_finish();
+	memdelete(scan_data->thread);
+	if (scan_progress) {
+		scan_progress->hide();
+	}
+
+	for (const String &E : scan_data->found_projects) {
+		add_project(E, false);
+	}
+	memdelete(scan_data);
+	scan_data = nullptr;
+
+	save_config();
+
+	if (ProjectManager::get_singleton()->is_initialized()) {
+		update_project_list();
 	}
 }
 
@@ -395,7 +558,7 @@ void ProjectList::_migrate_config() {
 		String path = EDITOR_GET(property_key);
 		print_line("Migrating legacy project '" + path + "'.");
 
-		String favoriteKey = "favorite_projects/" + property_key.get_slice("/", 1);
+		String favoriteKey = "favorite_projects/" + property_key.get_slicec('/', 1);
 		bool favorite = EditorSettings::get_singleton()->has_setting(favoriteKey);
 		add_project(path, favorite);
 		if (favorite) {
@@ -417,14 +580,16 @@ ProjectList::Item ProjectList::load_project_data(const String &p_path, bool p_fa
 	String conf = p_path.path_join("project.godot");
 	bool grayed = false;
 	bool missing = false;
+	bool recovery_mode = false;
 
 	Ref<ConfigFile> cf = memnew(ConfigFile);
 	Error cf_err = cf->load(conf);
 
 	int config_version = 0;
+	String cf_project_name;
 	String project_name = TTR("Unnamed Project");
 	if (cf_err == OK) {
-		String cf_project_name = cf->get_value("application", "config/name", "");
+		cf_project_name = cf->get_value("application", "config/name", "");
 		if (!cf_project_name.is_empty()) {
 			project_name = cf_project_name.xml_unescape();
 		}
@@ -438,9 +603,23 @@ ProjectList::Item ProjectList::load_project_data(const String &p_path, bool p_fa
 
 	const String description = cf->get_value("application", "config/description", "");
 	const PackedStringArray tags = cf->get_value("application", "config/tags", PackedStringArray());
-	const String icon = cf->get_value("application", "config/icon", "");
 	const String main_scene = cf->get_value("application", "run/main_scene", "");
 
+	String icon = cf->get_value("application", "config/icon", "");
+	if (icon.begins_with("uid://")) {
+		Error err;
+		Ref<FileAccess> file = FileAccess::open(p_path.path_join(".godot/uid_cache.bin"), FileAccess::READ, &err);
+		if (err == OK) {
+			icon = ResourceUID::get_path_from_cache(file, icon);
+			if (icon.is_empty()) {
+				WARN_PRINT(vformat("Could not load icon from UID for project at path \"%s\". Make sure UID cache exists.", p_path));
+			}
+		} else {
+			// Cache does not exist yet, so ignore and fallback to default icon.
+			icon = "";
+		}
+	}
+
 	PackedStringArray project_features = cf->get_value("application", "config/features", PackedStringArray());
 	PackedStringArray unsupported_features = ProjectSettings::get_unsupported_features(project_features);
 
@@ -458,7 +637,7 @@ ProjectList::Item ProjectList::load_project_data(const String &p_path, bool p_fa
 			unsupported_features.push_back("3.x");
 			project_version = "3.x";
 		} else {
-			unsupported_features.push_back("Unknown version");
+			unsupported_features.push_back(TTR("Unknown version"));
 		}
 	}
 
@@ -479,14 +658,35 @@ ProjectList::Item ProjectList::load_project_data(const String &p_path, bool p_fa
 	} else {
 		grayed = true;
 		missing = true;
-		print_line("Project is missing: " + conf);
 	}
 
 	for (const String &tag : tags) {
 		ProjectManager::get_singleton()->add_new_tag(tag);
 	}
 
-	return Item(project_name, description, project_version, tags, p_path, icon, main_scene, unsupported_features, last_edited, p_favorite, grayed, missing, config_version);
+	// We can't use OS::get_user_dir() because it attempts to load paths from the current loaded project through ProjectSettings,
+	// while here we're parsing project files externally. Therefore, we have to replicate its behavior.
+	String user_dir;
+	if (!cf_project_name.is_empty()) {
+		String appname = OS::get_singleton()->get_safe_dir_name(cf_project_name);
+		bool use_custom_dir = cf->get_value("application", "config/use_custom_user_dir", false);
+		if (use_custom_dir) {
+			String custom_dir = OS::get_singleton()->get_safe_dir_name(cf->get_value("application", "config/custom_user_dir_name", ""), true);
+			if (custom_dir.is_empty()) {
+				custom_dir = appname;
+			}
+			user_dir = custom_dir;
+		} else {
+			user_dir = OS::get_singleton()->get_godot_dir_name().path_join("app_userdata").path_join(appname);
+		}
+	} else {
+		user_dir = OS::get_singleton()->get_godot_dir_name().path_join("app_userdata").path_join("[unnamed project]");
+	}
+
+	String recovery_mode_lock_file = OS::get_singleton()->get_user_data_dir(user_dir).path_join(".recovery_mode_lock");
+	recovery_mode = FileAccess::exists(recovery_mode_lock_file);
+
+	return Item(project_name, description, project_version, tags, p_path, icon, main_scene, unsupported_features, last_edited, p_favorite, grayed, missing, recovery_mode, config_version);
 }
 
 void ProjectList::_update_icons_async() {
@@ -548,6 +748,7 @@ void ProjectList::update_project_list() {
 
 	set_v_scroll(0);
 	emit_signal(SNAME(SIGNAL_LIST_CHANGED));
+	queue_accessibility_update();
 }
 
 void ProjectList::sort_projects() {
@@ -564,7 +765,7 @@ void ProjectList::sort_projects() {
 			PackedStringArray remaining;
 			for (const String &part : search_parts) {
 				if (part.begins_with("tag:")) {
-					tags.push_back(part.get_slice(":", 1));
+					tags.push_back(part.get_slicec(':', 1));
 				} else {
 					remaining.append(part);
 				}
@@ -581,7 +782,7 @@ void ProjectList::sort_projects() {
 		bool item_visible = true;
 		if (!_search_term.is_empty()) {
 			String search_path;
-			if (search_term.contains("/")) {
+			if (search_term.contains_char('/')) {
 				// Search path will match the whole path
 				search_path = item.path;
 			} else {
@@ -612,6 +813,7 @@ void ProjectList::sort_projects() {
 	// Rewind the coroutine because order of projects changed
 	_update_icons_async();
 	update_dock_menu();
+	queue_accessibility_update();
 }
 
 int ProjectList::get_project_count() const {
@@ -624,31 +826,44 @@ void ProjectList::find_projects(const String &p_path) {
 }
 
 void ProjectList::find_projects_multiple(const PackedStringArray &p_paths) {
-	List<String> projects;
+	if (!scan_progress && is_inside_tree()) {
+		scan_progress = memnew(AcceptDialog);
+		scan_progress->set_title(TTRC("Scanning"));
+		scan_progress->set_ok_button_text(TTRC("Cancel"));
 
-	for (int i = 0; i < p_paths.size(); i++) {
-		const String &base_path = p_paths.get(i);
-		print_verbose(vformat("Scanning for projects in \"%s\".", base_path));
+		VBoxContainer *vb = memnew(VBoxContainer);
+		scan_progress->add_child(vb);
 
-		_scan_folder_recursive(base_path, &projects);
-		print_verbose(vformat("Found %d project(s).", projects.size()));
+		Label *label = memnew(Label);
+		label->set_text(TTRC("Scanning for projects..."));
+		vb->add_child(label);
+
+		ProgressBar *progress = memnew(ProgressBar);
+		progress->set_indeterminate(true);
+		vb->add_child(progress);
+
+		add_child(scan_progress);
+		scan_progress->connect(SceneStringName(confirmed), callable_mp(this, &ProjectList::_scan_finished));
+		scan_progress->connect("canceled", callable_mp(this, &ProjectList::_scan_finished));
 	}
 
-	for (const String &E : projects) {
-		add_project(E, false);
-	}
+	scan_data = memnew(ScanData);
+	scan_data->paths_to_scan = p_paths;
+	scan_data->scan_in_progress.set();
 
-	save_config();
+	scan_data->thread = memnew(Thread);
+	scan_data->thread->start(_scan_thread, scan_data);
 
-	if (ProjectManager::get_singleton()->is_initialized()) {
-		update_project_list();
+	if (scan_progress) {
+		scan_progress->reset_size();
+		scan_progress->popup_centered();
 	}
+	set_process(true);
 }
 
 void ProjectList::load_project_list() {
-	List<String> sections;
 	_config.load(_config_path);
-	_config.get_sections(&sections);
+	Vector<String> sections = _config.get_sections();
 
 	for (const String &path : sections) {
 		bool favorite = _config.get_value(path, "favorite", false);
@@ -656,7 +871,11 @@ void ProjectList::load_project_list() {
 	}
 }
 
-void ProjectList::_scan_folder_recursive(const String &p_path, List<String> *r_projects) {
+void ProjectList::_scan_folder_recursive(const String &p_path, List<String> *r_projects, const SafeFlag &p_scan_active) {
+	if (!p_scan_active.is_set()) {
+		return;
+	}
+
 	Ref<DirAccess> da = DirAccess::create(DirAccess::ACCESS_FILESYSTEM);
 	Error error = da->change_dir(p_path);
 	ERR_FAIL_COND_MSG(error != OK, vformat("Failed to open the path \"%s\" for scanning (code %d).", p_path, error));
@@ -664,8 +883,12 @@ void ProjectList::_scan_folder_recursive(const String &p_path, List<String> *r_p
 	da->list_dir_begin();
 	String n = da->get_next();
 	while (!n.is_empty()) {
+		if (!p_scan_active.is_set()) {
+			return;
+		}
+
 		if (da->current_is_dir() && n[0] != '.') {
-			_scan_folder_recursive(da->get_current_dir().path_join(n), r_projects);
+			_scan_folder_recursive(da->get_current_dir().path_join(n), r_projects, p_scan_active);
 		} else if (n == "project.godot") {
 			r_projects->push_back(da->get_current_dir());
 		}
@@ -680,6 +903,7 @@ void ProjectList::add_project(const String &dir_path, bool favorite) {
 	if (!_config.has_section(dir_path)) {
 		_config.set_value(dir_path, "favorite", favorite);
 	}
+	queue_accessibility_update();
 }
 
 void ProjectList::set_project_version(const String &p_project_path, int p_version) {
@@ -739,6 +963,15 @@ int ProjectList::refresh_project(const String &dir_path) {
 	return index;
 }
 
+int ProjectList::get_index(const ProjectListItemControl *p_control) const {
+	for (int i = 0; i < _projects.size(); ++i) {
+		if (_projects[i].control == p_control) {
+			return i;
+		}
+	}
+	return -1;
+}
+
 void ProjectList::ensure_project_visible(int p_index) {
 	const Item &item = _projects[p_index];
 	ensure_control_visible(item.control);
@@ -760,7 +993,7 @@ void ProjectList::_create_project_item_control(int p_index) {
 	hb->set_tags(item.tags, this);
 	hb->set_unsupported_features(item.unsupported_features.duplicate());
 	hb->set_project_version(item.project_version);
-	hb->set_last_edited_info(Time::get_singleton()->get_datetime_string_from_unix_time(item.last_edited, true));
+	hb->set_last_edited_info(item.get_last_edited_string());
 
 	hb->set_is_favorite(item.favorite);
 	hb->set_is_missing(item.missing);
@@ -806,6 +1039,7 @@ void ProjectList::_remove_project(int p_index, bool p_update_config) {
 		// Not actually saving the file, in case you are doing more changes to settings
 	}
 
+	queue_accessibility_update();
 	update_dock_menu();
 }
 
@@ -887,18 +1121,21 @@ void ProjectList::_clear_project_selection() {
 	for (int i = 0; i < previous_selected_items.size(); ++i) {
 		previous_selected_items[i].control->set_selected(false);
 	}
+	queue_accessibility_update();
 }
 
 void ProjectList::_select_project_nocheck(int p_index) {
 	Item &item = _projects.write[p_index];
 	_selected_project_paths.insert(item.path);
 	item.control->set_selected(true);
+	queue_accessibility_update();
 }
 
 void ProjectList::_deselect_project_nocheck(int p_index) {
 	Item &item = _projects.write[p_index];
 	_selected_project_paths.erase(item.path);
 	item.control->set_selected(false);
+	queue_accessibility_update();
 }
 
 inline void _sort_project_range(int &a, int &b) {
@@ -924,6 +1161,10 @@ void ProjectList::select_project(int p_index) {
 	_select_project_nocheck(p_index);
 }
 
+void ProjectList::deselect_project(int p_index) {
+	_deselect_project_nocheck(p_index);
+}
+
 void ProjectList::select_first_visible_project() {
 	_clear_project_selection();
 
@@ -937,7 +1178,7 @@ void ProjectList::select_first_visible_project() {
 
 Vector<ProjectList::Item> ProjectList::get_selected_projects() const {
 	Vector<Item> items;
-	if (_selected_project_paths.size() == 0) {
+	if (_selected_project_paths.is_empty()) {
 		return items;
 	}
 	items.resize(_selected_project_paths.size());
@@ -958,7 +1199,7 @@ const HashSet<String> &ProjectList::get_selected_project_keys() const {
 }
 
 int ProjectList::get_single_selected_index() const {
-	if (_selected_project_paths.size() == 0) {
+	if (_selected_project_paths.is_empty()) {
 		// Default selection
 		return 0;
 	}
@@ -979,7 +1220,7 @@ int ProjectList::get_single_selected_index() const {
 }
 
 void ProjectList::erase_selected_projects(bool p_delete_project_contents) {
-	if (_selected_project_paths.size() == 0) {
+	if (_selected_project_paths.is_empty()) {
 		return;
 	}
 
