@@ -31,16 +31,17 @@
 #include "base_button.h"
 
 #include "core/config/project_settings.h"
-#include "core/os/keyboard.h"
+#include "scene/gui/label.h"
 #include "scene/main/window.h"
 
 void BaseButton::_unpress_group() {
-	if (!button_group.is_valid()) {
+	if (button_group.is_null()) {
 		return;
 	}
 
 	if (toggle_mode && !button_group->is_allow_unpress()) {
 		status.pressed = true;
+		queue_accessibility_update();
 	}
 
 	for (BaseButton *E : button_group->buttons) {
@@ -83,15 +84,66 @@ void BaseButton::gui_input(const Ref<InputEvent> &p_event) {
 	}
 }
 
+void BaseButton::_accessibility_action_click(const Variant &p_data) {
+	if (toggle_mode) {
+		status.pressed = !status.pressed;
+
+		if (status.pressed) {
+			_unpress_group();
+			if (button_group.is_valid()) {
+				button_group->emit_signal(SceneStringName(pressed), this);
+			}
+		}
+
+		_toggled(status.pressed);
+		_pressed();
+	} else {
+		_pressed();
+	}
+	queue_accessibility_update();
+	queue_redraw();
+}
+
 void BaseButton::_notification(int p_what) {
 	switch (p_what) {
+		case NOTIFICATION_ACCESSIBILITY_UPDATE: {
+			RID ae = get_accessibility_element();
+			ERR_FAIL_COND(ae.is_null());
+
+			DisplayServer::get_singleton()->accessibility_update_set_role(ae, DisplayServer::AccessibilityRole::ROLE_BUTTON);
+
+			DisplayServer::get_singleton()->accessibility_update_add_action(ae, DisplayServer::AccessibilityAction::ACTION_CLICK, callable_mp(this, &BaseButton::_accessibility_action_click));
+			DisplayServer::get_singleton()->accessibility_update_set_flag(ae, DisplayServer::AccessibilityFlags::FLAG_DISABLED, status.disabled);
+			if (toggle_mode) {
+				DisplayServer::get_singleton()->accessibility_update_set_checked(ae, status.pressed);
+			}
+			if (button_group.is_valid()) {
+				for (const BaseButton *btn : button_group->buttons) {
+					if (btn->is_part_of_edited_scene()) {
+						continue;
+					}
+					DisplayServer::get_singleton()->accessibility_update_add_related_radio_group(ae, btn->get_accessibility_element());
+				}
+			}
+			if (shortcut_in_tooltip && shortcut.is_valid() && shortcut->has_valid_event()) {
+				String text = atr(shortcut->get_name()) + " (" + shortcut->get_as_text() + ")";
+				String tooltip = get_tooltip_text();
+				if (!tooltip.is_empty() && shortcut->get_name().nocasecmp_to(tooltip) != 0) {
+					text += "\n" + atr(tooltip);
+				}
+				DisplayServer::get_singleton()->accessibility_update_set_tooltip(ae, text);
+			}
+		} break;
+
 		case NOTIFICATION_MOUSE_ENTER: {
 			status.hovering = true;
+			queue_accessibility_update();
 			queue_redraw();
 		} break;
 
 		case NOTIFICATION_MOUSE_EXIT: {
 			status.hovering = false;
+			queue_accessibility_update();
 			queue_redraw();
 		} break;
 
@@ -114,6 +166,11 @@ void BaseButton::_notification(int p_what) {
 			} else if (status.hovering) {
 				queue_redraw();
 			}
+
+			if (status.pressed_down_with_focus) {
+				status.pressed_down_with_focus = false;
+				emit_signal(SNAME("button_up"));
+			}
 		} break;
 
 		case NOTIFICATION_VISIBILITY_CHANGED:
@@ -140,14 +197,19 @@ void BaseButton::_pressed() {
 void BaseButton::_toggled(bool p_pressed) {
 	GDVIRTUAL_CALL(_toggled, p_pressed);
 	toggled(p_pressed);
-	emit_signal(SNAME("toggled"), p_pressed);
+	emit_signal(SceneStringName(toggled), p_pressed);
 }
 
 void BaseButton::on_action_event(Ref<InputEvent> p_event) {
-	if (p_event->is_pressed()) {
+	Ref<InputEventMouseButton> mouse_button = p_event;
+
+	if (p_event->is_pressed() && (mouse_button.is_null() || status.hovering)) {
 		status.press_attempt = true;
 		status.pressing_inside = true;
-		emit_signal(SNAME("button_down"));
+		if (!status.pressed_down_with_focus) {
+			status.pressed_down_with_focus = true;
+			emit_signal(SNAME("button_down"));
+		}
 	}
 
 	if (status.press_attempt && status.pressing_inside) {
@@ -165,6 +227,7 @@ void BaseButton::on_action_event(Ref<InputEvent> p_event) {
 				}
 				_toggled(status.pressed);
 				_pressed();
+				queue_accessibility_update();
 			}
 		} else {
 			if ((p_event->is_pressed() && action_mode == ACTION_MODE_BUTTON_PRESS) || (!p_event->is_pressed() && action_mode == ACTION_MODE_BUTTON_RELEASE)) {
@@ -174,15 +237,12 @@ void BaseButton::on_action_event(Ref<InputEvent> p_event) {
 	}
 
 	if (!p_event->is_pressed()) {
-		Ref<InputEventMouseButton> mouse_button = p_event;
-		if (mouse_button.is_valid()) {
-			if (!has_point(mouse_button->get_position())) {
-				status.hovering = false;
-			}
-		}
 		status.press_attempt = false;
 		status.pressing_inside = false;
-		emit_signal(SNAME("button_up"));
+		if (status.pressed_down_with_focus) {
+			status.pressed_down_with_focus = false;
+			emit_signal(SNAME("button_up"));
+		}
 	}
 
 	queue_redraw();
@@ -206,8 +266,14 @@ void BaseButton::set_disabled(bool p_disabled) {
 		}
 		status.press_attempt = false;
 		status.pressing_inside = false;
+		if (status.pressed_down_with_focus) {
+			status.pressed_down_with_focus = false;
+			emit_signal(SNAME("button_up"));
+		}
 	}
+	queue_accessibility_update();
 	queue_redraw();
+	update_minimum_size();
 }
 
 bool BaseButton::is_disabled() const {
@@ -239,7 +305,7 @@ void BaseButton::set_pressed_no_signal(bool p_pressed) {
 		return;
 	}
 	status.pressed = p_pressed;
-
+	queue_accessibility_update();
 	queue_redraw();
 }
 
@@ -295,6 +361,7 @@ void BaseButton::set_toggle_mode(bool p_on) {
 	if (!p_on) {
 		set_pressed(false);
 	}
+	queue_accessibility_update();
 
 	toggle_mode = p_on;
 	update_configuration_warnings();
@@ -305,7 +372,10 @@ bool BaseButton::is_toggle_mode() const {
 }
 
 void BaseButton::set_shortcut_in_tooltip(bool p_on) {
-	shortcut_in_tooltip = p_on;
+	if (shortcut_in_tooltip != p_on) {
+		shortcut_in_tooltip = p_on;
+		queue_accessibility_update();
+	}
 }
 
 bool BaseButton::is_shortcut_in_tooltip_enabled() const {
@@ -345,8 +415,11 @@ bool BaseButton::is_shortcut_feedback() const {
 }
 
 void BaseButton::set_shortcut(const Ref<Shortcut> &p_shortcut) {
-	shortcut = p_shortcut;
-	set_process_shortcut_input(shortcut.is_valid());
+	if (shortcut != p_shortcut) {
+		shortcut = p_shortcut;
+		set_process_shortcut_input(shortcut.is_valid());
+		queue_accessibility_update();
+	}
 }
 
 Ref<Shortcut> BaseButton::get_shortcut() const {
@@ -372,7 +445,7 @@ void BaseButton::shortcut_input(const Ref<InputEvent> &p_event) {
 
 			_toggled(status.pressed);
 			_pressed();
-
+			queue_accessibility_update();
 		} else {
 			_pressed();
 		}
@@ -383,8 +456,8 @@ void BaseButton::shortcut_input(const Ref<InputEvent> &p_event) {
 			if (shortcut_feedback_timer == nullptr) {
 				shortcut_feedback_timer = memnew(Timer);
 				shortcut_feedback_timer->set_one_shot(true);
-				add_child(shortcut_feedback_timer);
-				shortcut_feedback_timer->set_wait_time(GLOBAL_GET("gui/timers/button_shortcut_feedback_highlight_time"));
+				add_child(shortcut_feedback_timer, false, INTERNAL_MODE_BACK);
+				shortcut_feedback_timer->set_wait_time(GLOBAL_GET_CACHED(double, "gui/timers/button_shortcut_feedback_highlight_time"));
 				shortcut_feedback_timer->connect("timeout", callable_mp(this, &BaseButton::_shortcut_feedback_timeout));
 			}
 
@@ -394,16 +467,31 @@ void BaseButton::shortcut_input(const Ref<InputEvent> &p_event) {
 	}
 }
 
-String BaseButton::get_tooltip(const Point2 &p_pos) const {
-	String tooltip = Control::get_tooltip(p_pos);
-	if (shortcut_in_tooltip && shortcut.is_valid() && shortcut->has_valid_event()) {
-		String text = shortcut->get_name() + " (" + shortcut->get_as_text() + ")";
-		if (!tooltip.is_empty() && shortcut->get_name().nocasecmp_to(tooltip) != 0) {
-			text += "\n" + atr(tooltip);
-		}
-		tooltip = text;
+Control *BaseButton::make_custom_tooltip(const String &p_text) const {
+	Control *control = Control::make_custom_tooltip(p_text);
+	if (control) {
+		return control;
 	}
-	return tooltip;
+	if (!shortcut_in_tooltip || shortcut.is_null() || !shortcut->has_valid_event()) {
+		return nullptr; // Use the default tooltip label.
+	}
+
+	String text = atr(shortcut->get_name()) + " (" + shortcut->get_as_text() + ")";
+	if (!p_text.is_empty() && shortcut->get_name().nocasecmp_to(p_text) != 0) {
+		text += "\n" + atr(p_text);
+	}
+
+	// Make a label similar to the default tooltip label.
+	// Auto translation is disabled because we already did that manually above.
+	//
+	// We can't customize the tooltip text by overriding `get_tooltip()`
+	// because otherwise user-defined `_make_custom_tooltip()` would receive
+	// the translated and annotated text.
+	Label *label = memnew(Label(text));
+	label->set_auto_translate_mode(AUTO_TRANSLATE_MODE_DISABLED);
+	label->set_theme_type_variation(SNAME("TooltipLabel"));
+
+	return label;
 }
 
 void BaseButton::set_button_group(const Ref<ButtonGroup> &p_group) {
@@ -417,6 +505,7 @@ void BaseButton::set_button_group(const Ref<ButtonGroup> &p_group) {
 		button_group->buttons.insert(this);
 	}
 
+	queue_accessibility_update();
 	queue_redraw(); //checkbox changes to radio if set a buttongroup
 	update_configuration_warnings();
 }
