@@ -32,10 +32,20 @@
 
 #include "scene/theme/theme_db.h"
 
+Rect2 TabContainer::_get_tab_rect() const {
+	Rect2 rect;
+	if (tabs_visible && get_tab_count() > 0) {
+		rect = Rect2(theme_cache.tabbar_style->get_offset(), tab_bar->get_size());
+		rect.position.x += is_layout_rtl() ? theme_cache.menu_icon->get_width() : theme_cache.side_margin;
+	}
+
+	return rect;
+}
+
 int TabContainer::_get_tab_height() const {
 	int height = 0;
 	if (tabs_visible && get_tab_count() > 0) {
-		height = tab_bar->get_minimum_size().height;
+		height = tab_bar->get_minimum_size().height + theme_cache.tabbar_style->get_margin(SIDE_TOP) + theme_cache.tabbar_style->get_margin(SIDE_BOTTOM);
 	}
 
 	return height;
@@ -50,29 +60,27 @@ void TabContainer::gui_input(const Ref<InputEvent> &p_event) {
 
 	if (mb.is_valid() && mb->is_pressed() && mb->get_button_index() == MouseButton::LEFT) {
 		Point2 pos = mb->get_position();
-		Size2 size = get_size();
-		real_t content_height = size.height - _get_tab_height();
+		real_t content_height = get_size().height - _get_tab_height();
+
+		Rect2 popup_rect = _get_tab_rect();
+		popup_rect.position.x += is_layout_rtl() ? -theme_cache.menu_icon->get_width() : popup_rect.size.x;
+		popup_rect.position.y += tabs_position == POSITION_BOTTOM ? content_height : 0;
+		popup_rect.size.x = theme_cache.menu_icon->get_width();
 
 		// Click must be on tabs in the tab header area.
-		if (tabs_position == POSITION_TOP && pos.y > _get_tab_height()) {
-			return;
-		}
-		if (tabs_position == POSITION_BOTTOM && pos.y < content_height) {
+		if (!tabs_visible || pos.y < popup_rect.position.y || pos.y >= popup_rect.position.y + popup_rect.size.y) {
 			return;
 		}
 
 		// Handle menu button.
 		if (popup) {
-			if (is_layout_rtl() ? pos.x < theme_cache.menu_icon->get_width() : pos.x > size.width - theme_cache.menu_icon->get_width()) {
+			if (popup_rect.has_point(pos)) {
 				emit_signal(SNAME("pre_popup_pressed"));
 
 				Vector2 popup_pos = get_screen_position();
-				if (!is_layout_rtl()) {
-					popup_pos.x += size.width - popup->get_size().width;
-				}
-				popup_pos.y += _get_tab_height() / 2.0;
+				popup_pos.x += popup_rect.position.x + (is_layout_rtl() ? 0 : popup_rect.size.x - popup->get_size().width);
+				popup_pos.y += popup_rect.position.y + popup_rect.size.y / 2.0;
 				if (tabs_position == POSITION_BOTTOM) {
-					popup_pos.y += content_height;
 					popup_pos.y -= popup->get_size().height;
 					popup_pos.y -= theme_cache.menu_icon->get_height() / 2.0;
 				} else {
@@ -90,17 +98,15 @@ void TabContainer::gui_input(const Ref<InputEvent> &p_event) {
 
 	if (mm.is_valid()) {
 		Point2 pos = mm->get_position();
-		Size2 size = get_size();
+		real_t content_height = get_size().height - _get_tab_height();
+
+		Rect2 popup_rect = _get_tab_rect();
+		popup_rect.position.x += is_layout_rtl() ? -theme_cache.menu_icon->get_width() : popup_rect.size.x;
+		popup_rect.position.y += tabs_position == POSITION_BOTTOM ? content_height : 0;
+		popup_rect.size.x = theme_cache.menu_icon->get_width();
 
 		// Mouse must be on tabs in the tab header area.
-		if (tabs_position == POSITION_TOP && pos.y > _get_tab_height()) {
-			if (menu_hovered) {
-				menu_hovered = false;
-				queue_redraw();
-			}
-			return;
-		}
-		if (tabs_position == POSITION_BOTTOM && pos.y < size.height - _get_tab_height()) {
+		if (!tabs_visible || pos.y < popup_rect.position.y || pos.y >= popup_rect.position.y + popup_rect.size.y) {
 			if (menu_hovered) {
 				menu_hovered = false;
 				queue_redraw();
@@ -109,28 +115,15 @@ void TabContainer::gui_input(const Ref<InputEvent> &p_event) {
 		}
 
 		if (popup) {
-			if (is_layout_rtl()) {
-				if (pos.x <= theme_cache.menu_icon->get_width()) {
-					if (!menu_hovered) {
-						menu_hovered = true;
-						queue_redraw();
-						return;
-					}
-				} else if (menu_hovered) {
-					menu_hovered = false;
-					queue_redraw();
-				}
-			} else {
-				if (pos.x >= size.width - theme_cache.menu_icon->get_width()) {
-					if (!menu_hovered) {
-						menu_hovered = true;
-						queue_redraw();
-						return;
-					}
-				} else if (menu_hovered) {
-					menu_hovered = false;
+			if (popup_rect.has_point(pos)) {
+				if (!menu_hovered) {
+					menu_hovered = true;
 					queue_redraw();
+					return;
 				}
+			} else if (menu_hovered) {
+				menu_hovered = false;
+				queue_redraw();
 			}
 
 			if (menu_hovered) {
@@ -142,6 +135,45 @@ void TabContainer::gui_input(const Ref<InputEvent> &p_event) {
 
 void TabContainer::_notification(int p_what) {
 	switch (p_what) {
+		case NOTIFICATION_ACCESSIBILITY_INVALIDATE: {
+			tab_panels.clear();
+		} break;
+
+		case NOTIFICATION_ACCESSIBILITY_UPDATE: {
+			RID ae = get_accessibility_element();
+			ERR_FAIL_COND(ae.is_null());
+
+			int tab_index = 0;
+			int tab_cur = tab_bar->get_current_tab();
+			for (int i = 0; i < get_child_count(); i++) {
+				Node *child_node = get_child(i);
+				Window *child_wnd = Object::cast_to<Window>(child_node);
+				if (child_wnd && !child_wnd->is_embedded()) {
+					continue;
+				}
+				if (child_node->is_part_of_edited_scene()) {
+					continue;
+				}
+				Control *control = as_sortable_control(child_node, SortableVisibilityMode::IGNORE);
+				if (!control || control == tab_bar || children_removing.has(control)) {
+					DisplayServer::get_singleton()->accessibility_update_add_child(ae, child_node->get_accessibility_element());
+				} else {
+					if (!tab_panels.has(child_node)) {
+						tab_panels[child_node] = DisplayServer::get_singleton()->accessibility_create_sub_element(ae, DisplayServer::AccessibilityRole::ROLE_TAB_PANEL);
+					}
+					RID panel = tab_panels[child_node];
+					RID tab = tab_bar->get_tab_accessibility_element(tab_index);
+
+					DisplayServer::get_singleton()->accessibility_update_add_related_controls(tab, panel);
+					DisplayServer::get_singleton()->accessibility_update_add_related_labeled_by(panel, tab);
+					DisplayServer::get_singleton()->accessibility_update_set_flag(panel, DisplayServer::AccessibilityFlags::FLAG_HIDDEN, tab_index != tab_cur);
+					DisplayServer::get_singleton()->accessibility_update_add_child(panel, child_node->get_accessibility_element());
+
+					tab_index++;
+				}
+			}
+		} break;
+
 		case NOTIFICATION_ENTER_TREE: {
 			// If some nodes happen to be renamed outside the tree, the tab names need to be updated manually.
 			if (get_tab_count() > 0) {
@@ -162,6 +194,7 @@ void TabContainer::_notification(int p_what) {
 		case NOTIFICATION_DRAW: {
 			RID canvas = get_canvas_item();
 			Size2 size = get_size();
+			Rect2 tabbar_rect = _get_tab_rect();
 
 			// Draw only the tab area if the header is hidden.
 			if (!tabs_visible) {
@@ -179,18 +212,19 @@ void TabContainer::_notification(int p_what) {
 
 			// Draw the popup menu.
 			if (get_popup()) {
-				int x = is_layout_rtl() ? 0 : get_size().width - theme_cache.menu_icon->get_width();
+				int x = is_layout_rtl() ? tabbar_rect.position.x - theme_cache.menu_icon->get_width() : tabbar_rect.position.x + tabbar_rect.size.x;
+				header_voffset += tabbar_rect.position.y;
 
 				if (menu_hovered) {
-					theme_cache.menu_hl_icon->draw(get_canvas_item(), Point2(x, header_voffset + (header_height - theme_cache.menu_hl_icon->get_height()) / 2));
+					theme_cache.menu_hl_icon->draw(get_canvas_item(), Point2(x, header_voffset + (tabbar_rect.size.y - theme_cache.menu_hl_icon->get_height()) / 2));
 				} else {
-					theme_cache.menu_icon->draw(get_canvas_item(), Point2(x, header_voffset + (header_height - theme_cache.menu_icon->get_height()) / 2));
+					theme_cache.menu_icon->draw(get_canvas_item(), Point2(x, header_voffset + (tabbar_rect.size.y - theme_cache.menu_icon->get_height()) / 2));
 				}
 			}
 		} break;
 
 		case NOTIFICATION_VISIBILITY_CHANGED: {
-			if (!is_visible() || setup_current_tab > -2) {
+			if (!is_visible()) {
 				return;
 			}
 
@@ -200,7 +234,7 @@ void TabContainer::_notification(int p_what) {
 			// beat it to the punch and make sure that the correct node is the only one visible first.
 			// Otherwise, it can prevent a tab change done right before this container was made visible.
 			Vector<Control *> controls = _get_tab_controls();
-			int current = get_current_tab();
+			int current = setup_current_tab > -2 ? setup_current_tab : get_current_tab();
 			for (int i = 0; i < controls.size(); i++) {
 				controls[i]->set_visible(i == current);
 			}
@@ -247,6 +281,7 @@ void TabContainer::_on_theme_changed() {
 	tab_bar->add_theme_font_size_override(SceneStringName(font_size), theme_cache.tab_font_size);
 
 	tab_bar->add_theme_constant_override(SNAME("h_separation"), theme_cache.icon_separation);
+	tab_bar->add_theme_constant_override(SNAME("tab_separation"), theme_cache.tab_separation);
 	tab_bar->add_theme_constant_override(SNAME("icon_max_width"), theme_cache.icon_max_width);
 	tab_bar->add_theme_constant_override(SNAME("outline_size"), theme_cache.outline_size);
 
@@ -267,10 +302,17 @@ void TabContainer::_repaint() {
 	Vector<Control *> controls = _get_tab_controls();
 	int current = get_current_tab();
 
+	float top_margin = theme_cache.tabbar_style->get_margin(SIDE_TOP);
+	float bottom_margin = theme_cache.tabbar_style->get_margin(SIDE_BOTTOM);
+
+	// Move the TabBar to the top or bottom.
+	// Don't change the left and right offsets since the TabBar will resize and may change tab offset.
 	if (tabs_position == POSITION_BOTTOM) {
-		tab_bar->set_anchors_and_offsets_preset(PRESET_BOTTOM_WIDE);
+		tab_bar->set_anchor_and_offset(SIDE_BOTTOM, 1.0, -bottom_margin);
+		tab_bar->set_anchor_and_offset(SIDE_TOP, 1.0, top_margin - _get_tab_height());
 	} else {
-		tab_bar->set_anchors_and_offsets_preset(PRESET_TOP_WIDE);
+		tab_bar->set_anchor_and_offset(SIDE_TOP, 0.0, top_margin);
+		tab_bar->set_anchor_and_offset(SIDE_BOTTOM, 0.0, _get_tab_height() - bottom_margin);
 	}
 
 	updating_visibility = true;
@@ -299,51 +341,58 @@ void TabContainer::_repaint() {
 	}
 	updating_visibility = false;
 
-	_update_margins();
 	update_minimum_size();
 }
 
 void TabContainer::_update_margins() {
-	int menu_width = theme_cache.menu_icon->get_width();
-
 	// Directly check for validity, to avoid errors when quitting.
 	bool has_popup = popup_obj_id.is_valid();
 
-	if (get_tab_count() == 0) {
-		tab_bar->set_offset(SIDE_LEFT, 0);
-		tab_bar->set_offset(SIDE_RIGHT, has_popup ? -menu_width : 0);
+	int left_margin = theme_cache.tabbar_style->get_margin(SIDE_LEFT);
+	int right_margin = theme_cache.tabbar_style->get_margin(SIDE_RIGHT);
 
+	if (is_layout_rtl()) {
+		SWAP(left_margin, right_margin);
+	}
+
+	if (has_popup) {
+		right_margin += theme_cache.menu_icon->get_width();
+	}
+
+	if (get_tab_count() == 0) {
+		tab_bar->set_offset(SIDE_LEFT, left_margin);
+		tab_bar->set_offset(SIDE_RIGHT, -right_margin);
 		return;
 	}
 
 	switch (get_tab_alignment()) {
 		case TabBar::ALIGNMENT_LEFT: {
-			tab_bar->set_offset(SIDE_LEFT, theme_cache.side_margin);
-			tab_bar->set_offset(SIDE_RIGHT, has_popup ? -menu_width : 0);
+			tab_bar->set_offset(SIDE_LEFT, left_margin + theme_cache.side_margin);
+			tab_bar->set_offset(SIDE_RIGHT, -right_margin);
 		} break;
 
 		case TabBar::ALIGNMENT_CENTER: {
-			tab_bar->set_offset(SIDE_LEFT, 0);
-			tab_bar->set_offset(SIDE_RIGHT, has_popup ? -menu_width : 0);
+			tab_bar->set_offset(SIDE_LEFT, left_margin);
+			tab_bar->set_offset(SIDE_RIGHT, -right_margin);
 		} break;
 
 		case TabBar::ALIGNMENT_RIGHT: {
-			tab_bar->set_offset(SIDE_LEFT, 0);
+			tab_bar->set_offset(SIDE_LEFT, left_margin);
 
 			if (has_popup) {
-				tab_bar->set_offset(SIDE_RIGHT, -menu_width);
+				tab_bar->set_offset(SIDE_RIGHT, -right_margin);
 				return;
 			}
 
 			int first_tab_pos = tab_bar->get_tab_rect(0).position.x;
 			Rect2 last_tab_rect = tab_bar->get_tab_rect(get_tab_count() - 1);
-			int total_tabs_width = last_tab_rect.position.x - first_tab_pos + last_tab_rect.size.width;
+			int total_tabs_width = left_margin + right_margin + last_tab_rect.position.x - first_tab_pos + last_tab_rect.size.width;
 
 			// Calculate if all the tabs would still fit if the margin was present.
 			if (get_clip_tabs() && (tab_bar->get_offset_buttons_visible() || (get_tab_count() > 1 && (total_tabs_width + theme_cache.side_margin) > get_size().width))) {
-				tab_bar->set_offset(SIDE_RIGHT, has_popup ? -menu_width : 0);
+				tab_bar->set_offset(SIDE_RIGHT, -right_margin);
 			} else {
-				tab_bar->set_offset(SIDE_RIGHT, -theme_cache.side_margin);
+				tab_bar->set_offset(SIDE_RIGHT, -right_margin - theme_cache.side_margin);
 			}
 		} break;
 
@@ -362,7 +411,7 @@ void TabContainer::_on_mouse_exited() {
 Vector<Control *> TabContainer::_get_tab_controls() const {
 	Vector<Control *> controls;
 	for (int i = 0; i < get_child_count(); i++) {
-		Control *control = as_sortable_control(get_child(i), SortableVisbilityMode::IGNORE);
+		Control *control = as_sortable_control(get_child(i), SortableVisibilityMode::IGNORE);
 		if (!control || control == tab_bar || children_removing.has(control)) {
 			continue;
 		}
@@ -539,7 +588,7 @@ void TabContainer::add_child_notify(Node *p_child) {
 		return;
 	}
 
-	Control *c = as_sortable_control(p_child, SortableVisbilityMode::IGNORE);
+	Control *c = as_sortable_control(p_child, SortableVisibilityMode::IGNORE);
 	if (!c) {
 		return;
 	}
@@ -552,6 +601,7 @@ void TabContainer::add_child_notify(Node *p_child) {
 	if (get_tab_count() == 1) {
 		queue_redraw();
 	}
+	queue_accessibility_update();
 
 	p_child->connect("renamed", callable_mp(this, &TabContainer::_refresh_tab_names));
 	p_child->connect(SceneStringName(visibility_changed), callable_mp(this, &TabContainer::_on_tab_visibility_changed).bind(c));
@@ -569,22 +619,28 @@ void TabContainer::move_child_notify(Node *p_child) {
 		return;
 	}
 
-	Control *c = as_sortable_control(p_child, SortableVisbilityMode::IGNORE);
+	Control *c = as_sortable_control(p_child, SortableVisibilityMode::IGNORE);
 	if (c) {
 		tab_bar->move_tab(c->get_meta("_tab_index"), get_tab_idx_from_control(c));
 	}
 
 	_refresh_tab_indices();
+	queue_accessibility_update();
 }
 
 void TabContainer::remove_child_notify(Node *p_child) {
 	Container::remove_child_notify(p_child);
 
+	if (tab_panels.has(p_child)) {
+		DisplayServer::get_singleton()->accessibility_free_element(tab_panels[p_child]);
+		tab_panels.erase(p_child);
+	}
+
 	if (p_child == tab_bar) {
 		return;
 	}
 
-	Control *c = as_sortable_control(p_child, SortableVisbilityMode::IGNORE);
+	Control *c = as_sortable_control(p_child, SortableVisibilityMode::IGNORE);
 	if (!c) {
 		return;
 	}
@@ -603,6 +659,7 @@ void TabContainer::remove_child_notify(Node *p_child) {
 	if (get_tab_count() == 0) {
 		queue_redraw();
 	}
+	queue_accessibility_update();
 
 	p_child->remove_meta("_tab_index");
 	p_child->remove_meta("_tab_name");
@@ -891,15 +948,17 @@ Size2 TabContainer::get_minimum_size() const {
 
 	if (tabs_visible) {
 		ms = tab_bar->get_minimum_size();
+		ms.width += theme_cache.tabbar_style->get_margin(SIDE_LEFT) + theme_cache.tabbar_style->get_margin(SIDE_RIGHT);
+		ms.height += theme_cache.tabbar_style->get_margin(SIDE_TOP) + theme_cache.tabbar_style->get_margin(SIDE_BOTTOM);
 
 		if (!get_clip_tabs()) {
 			if (get_popup()) {
-				ms.x += theme_cache.menu_icon->get_width();
+				ms.width += theme_cache.menu_icon->get_width();
 			}
 
 			if (theme_cache.side_margin > 0 && get_tab_alignment() != TabBar::ALIGNMENT_CENTER &&
 					(get_tab_alignment() != TabBar::ALIGNMENT_RIGHT || !get_popup())) {
-				ms.x += theme_cache.side_margin;
+				ms.width += theme_cache.side_margin;
 			}
 		}
 	}
@@ -909,19 +968,19 @@ Size2 TabContainer::get_minimum_size() const {
 	for (int i = 0; i < controls.size(); i++) {
 		Control *c = controls[i];
 
-		if (!c->is_visible_in_tree() && !use_hidden_tabs_for_min_size) {
+		if (!c->is_visible() && !use_hidden_tabs_for_min_size) {
 			continue;
 		}
 
 		Size2 cms = c->get_combined_minimum_size();
 		largest_child_min_size = largest_child_min_size.max(cms);
 	}
-	ms.y += largest_child_min_size.y;
+	ms.height += largest_child_min_size.height;
 
 	Size2 panel_ms = theme_cache.panel_style->get_minimum_size();
 
-	ms.x = MAX(ms.x, largest_child_min_size.x + panel_ms.x);
-	ms.y += panel_ms.y;
+	ms.width = MAX(ms.width, largest_child_min_size.width + panel_ms.width);
+	ms.height += panel_ms.height;
 
 	return ms;
 }
@@ -947,7 +1006,7 @@ void TabContainer::set_popup(Node *p_popup) {
 
 Popup *TabContainer::get_popup() const {
 	if (popup_obj_id.is_valid()) {
-		Popup *popup = Object::cast_to<Popup>(ObjectDB::get_instance(popup_obj_id));
+		Popup *popup = ObjectDB::get_instance<Popup>(popup_obj_id);
 		if (popup) {
 			return popup;
 		} else {
@@ -1069,11 +1128,14 @@ void TabContainer::_bind_methods() {
 	ADD_PROPERTY(PropertyInfo(Variant::INT, "tab_focus_mode", PROPERTY_HINT_ENUM, "None,Click,All"), "set_tab_focus_mode", "get_tab_focus_mode");
 	ADD_PROPERTY(PropertyInfo(Variant::BOOL, "deselect_enabled"), "set_deselect_enabled", "get_deselect_enabled");
 
+	ADD_CLASS_DEPENDENCY("TabBar");
+
 	BIND_ENUM_CONSTANT(POSITION_TOP);
 	BIND_ENUM_CONSTANT(POSITION_BOTTOM);
 	BIND_ENUM_CONSTANT(POSITION_MAX);
 
 	BIND_THEME_ITEM(Theme::DATA_TYPE_CONSTANT, TabContainer, side_margin);
+	BIND_THEME_ITEM(Theme::DATA_TYPE_CONSTANT, TabContainer, tab_separation);
 
 	BIND_THEME_ITEM_CUSTOM(Theme::DATA_TYPE_STYLEBOX, TabContainer, panel_style, "panel");
 	BIND_THEME_ITEM_CUSTOM(Theme::DATA_TYPE_STYLEBOX, TabContainer, tabbar_style, "tabbar_background");
@@ -1113,6 +1175,7 @@ TabContainer::TabContainer() {
 	tab_bar = memnew(TabBar);
 	SET_DRAG_FORWARDING_GCDU(tab_bar, TabContainer);
 	add_child(tab_bar, false, INTERNAL_MODE_FRONT);
+	tab_bar->set_use_parent_material(true);
 	tab_bar->set_anchors_and_offsets_preset(Control::PRESET_TOP_WIDE);
 	tab_bar->connect("tab_changed", callable_mp(this, &TabContainer::_on_tab_changed));
 	tab_bar->connect("tab_clicked", callable_mp(this, &TabContainer::_on_tab_clicked));
