@@ -28,13 +28,14 @@
 /* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                 */
 /**************************************************************************/
 
-#ifndef PHYSICS_SERVER_2D_H
-#define PHYSICS_SERVER_2D_H
+#pragma once
 
 #include "core/io/resource.h"
 #include "core/object/class_db.h"
 #include "core/object/ref_counted.h"
 
+constexpr int MAX_CONTACTS_REPORTED_2D_MAX = 4096;
+
 class PhysicsDirectSpaceState2D;
 template <typename T>
 class TypedArray;
@@ -87,6 +88,12 @@ public:
 	virtual void set_sleep_state(bool p_enable) = 0;
 	virtual bool is_sleeping() const = 0;
 
+	virtual void set_collision_layer(uint32_t p_layer) = 0;
+	virtual uint32_t get_collision_layer() const = 0;
+
+	virtual void set_collision_mask(uint32_t p_mask) = 0;
+	virtual uint32_t get_collision_mask() const = 0;
+
 	virtual int get_contact_count() const = 0;
 
 	virtual Vector2 get_contact_local_position(int p_contact_idx) const = 0;
@@ -845,5 +852,3 @@ VARIANT_ENUM_CAST(PhysicsServer2D::PinJointFlag);
 VARIANT_ENUM_CAST(PhysicsServer2D::DampedSpringParam);
 VARIANT_ENUM_CAST(PhysicsServer2D::AreaBodyStatus);
 VARIANT_ENUM_CAST(PhysicsServer2D::ProcessInfo);
-
-#endif // PHYSICS_SERVER_2D_H
