@@ -28,8 +28,7 @@
 /* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                 */
 /**************************************************************************/
 
-#ifndef DEBUG_EFFECTS_RD_H
-#define DEBUG_EFFECTS_RD_H
+#pragma once
 
 #include "servers/rendering/renderer_rd/pipeline_cache_rd.h"
 #include "servers/rendering/renderer_rd/shaders/effects/motion_vectors.glsl.gen.h"
@@ -94,5 +93,3 @@ public:
 };
 
 } // namespace RendererRD
-
-#endif // DEBUG_EFFECTS_RD_H
