@@ -28,8 +28,7 @@
 /* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                 */
 /**************************************************************************/
 
-#ifndef RENDER_SCENE_DATA_RD_H
-#define RENDER_SCENE_DATA_RD_H
+#pragma once
 
 #include "servers/rendering/renderer_scene_render.h"
 #include "servers/rendering/storage/render_scene_data.h"
@@ -76,6 +75,7 @@ public:
 	float opaque_prepass_threshold = 0.0;
 	bool material_uv2_mode = false;
 	float emissive_exposure_normalization = 0.0;
+	bool shadow_pass = false;
 
 	Size2 shadow_atlas_pixel_size;
 	Size2 directional_shadow_pixel_size;
@@ -97,6 +97,18 @@ public:
 private:
 	RID uniform_buffer; // loaded into this uniform buffer (supplied externally)
 
+	enum SceneDataFlags {
+		SCENE_DATA_FLAGS_USE_AMBIENT_LIGHT = 1 << 0,
+		SCENE_DATA_FLAGS_USE_AMBIENT_CUBEMAP = 1 << 1,
+		SCENE_DATA_FLAGS_USE_REFLECTION_CUBEMAP = 1 << 2,
+		SCENE_DATA_FLAGS_USE_ROUGHNESS_LIMITER = 1 << 3,
+		SCENE_DATA_FLAGS_USE_FOG = 1 << 4,
+		SCENE_DATA_FLAGS_USE_UV2_MATERIAL = 1 << 5,
+		SCENE_DATA_FLAGS_USE_PANCAKE_SHADOWS = 1 << 6,
+		SCENE_DATA_FLAGS_IN_SHADOW_PASS = 1 << 7, // Only used by Forward+ renderer.
+		SCENE_DATA_FLAGS_MAX
+	};
+
 	// This struct is loaded into Set 1 - Binding 0, populated at start of rendering a frame, must match with shader code
 	struct UBO {
 		float projection_matrix[16];
@@ -118,15 +130,6 @@ private:
 		float penumbra_shadow_kernel[128];
 		float soft_shadow_kernel[128];
 
-		float radiance_inverse_xform[12];
-
-		float ambient_light_color_energy[4];
-
-		float ambient_color_sky_mix;
-		uint32_t use_ambient_light;
-		uint32_t use_ambient_cubemap;
-		uint32_t use_reflection_cubemap;
-
 		float shadow_atlas_pixel_size[2];
 		float directional_shadow_pixel_size[2];
 
@@ -135,36 +138,34 @@ private:
 		float z_far;
 		float z_near;
 
-		uint32_t roughness_limiter_enabled;
 		float roughness_limiter_amount;
 		float roughness_limiter_limit;
 		float opaque_prepass_threshold;
+		uint32_t flags;
 
-		// Fog
-		uint32_t fog_enabled;
-		uint32_t fog_mode;
+		float radiance_inverse_xform[12];
+
+		float ambient_light_color_energy[4];
+
+		float ambient_color_sky_mix;
 		float fog_density;
 		float fog_height;
-
 		float fog_height_density;
+
 		float fog_depth_curve;
 		float fog_depth_begin;
-		float taa_frame_count; // Used to add break up samples over multiple frames. Value is an integer from 0 to taa_phase_count -1.
+		float fog_depth_end;
+		float fog_sun_scatter;
 
 		float fog_light_color[3];
-		float fog_depth_end;
-
-		float fog_sun_scatter;
 		float fog_aerial_perspective;
+
 		float time;
-		float reflection_multiplier;
-
+		float taa_frame_count; // Used to add break up samples over multiple frames. Value is an integer from 0 to taa_phase_count -1.
 		float taa_jitter[2];
-		uint32_t material_uv2_mode;
-		float emissive_exposure_normalization; // Needed to normalize emissive when using physical units.
 
+		float emissive_exposure_normalization; // Needed to normalize emissive when using physical units.
 		float IBL_exposure_normalization; // Adjusts for baked exposure.
-		uint32_t pancake_shadows;
 		uint32_t camera_visible_layers;
 		float pass_alpha_multiplier;
 	};
@@ -174,5 +175,3 @@ private:
 		UBO prev_ubo;
 	};
 };
-
-#endif // RENDER_SCENE_DATA_RD_H
